#include "tcp-dtn.h"

BundleManagerTCP::BundleManagerTCP()
{
	pHead_ = NULL;
}

BundleManagerTCP::~BundleManagerTCP()
{
	EmptyList();
}

/* Add bundle data to the queue in TCP sequence number order. */
void BundleManagerTCP::AddBundle(Bundle newBundle)
{	
	newBundle.NextBundle() = NULL;
	// printf("New bundle seqNumber = %d\n", newBundle.TCPSequenceNo());
	
	if (!pHead_)
	{
		Bundle *pBundle = new Bundle(); *pBundle = newBundle;
		pHead_ = pBundle;
		// printf("Adding head seqNumber = %d\n", pBundle->TCPSequenceNo());
		return;
	}
	
	if (pHead_->TCPSequenceNo() > newBundle.TCPSequenceNo())
	{
		Bundle *pBundle = new Bundle(); *pBundle = newBundle;
		pBundle->NextBundle() = pHead_;
		pHead_ = pBundle;
		// printf("Adding first seqNumber = %d\n", pBundle->TCPSequenceNo());
		return;
	}
	
	Bundle *pLoopBundle = pHead_;
	
	while (pLoopBundle->NextBundle() && 
			(pLoopBundle->NextBundle()->TCPSequenceNo() < newBundle.TCPSequenceNo())
			)
	{
		pLoopBundle = pLoopBundle->NextBundle();
	}
	
	/* If not duplicate TCP packet received */
	if (pLoopBundle->TCPSequenceNo() != newBundle.TCPSequenceNo())
	{
		Bundle *pBundle = new Bundle(); *pBundle = newBundle;
		pBundle->NextBundle() = pLoopBundle->NextBundle();
		pLoopBundle->NextBundle() = pBundle;
	}
	// else
		// printf("Ignoring seqNumber = %d\n", newBundle.TCPSequenceNo());
}

/* Get corresponding bundle data for a certain TCP sequence number. */
Bundle* BundleManagerTCP::GetBundle(int nSequenceNo)
{
	Bundle *pLoopBundle = pHead_;
	
	while (pLoopBundle)
	{
		// printf("seqNumber = %d, runningData = %d, nextRunningData = %d\n", pRunningBundleData->m_nTCPSequenceNumber, 
		//	pRunningBundleData, pRunningBundleData->m_pNextBundle);
		if (nSequenceNo == pLoopBundle->TCPSequenceNo())
			return pLoopBundle;
		else
			pLoopBundle = pLoopBundle->NextBundle();
	}
	return NULL;
}

/* Get first corresponding bundle data of a certain bundle. */
Bundle* BundleManagerTCP::FindFirstBundle(Bundle inBundle)
{
	Bundle *pLoopBundle = pHead_;
	
	while (pLoopBundle)
	{
		// printf("seqNumber = %d, runningData = %d, nextRunningData = %d\n", pRunningBundleData->m_nTCPSequenceNumber, 
		//	pRunningBundleData, pRunningBundleData->m_pNextBundle);
		if (pLoopBundle->IsSameAs(inBundle))
			return pLoopBundle;
		else
			pLoopBundle = pLoopBundle->NextBundle();
	}
	return NULL;
}

/* 
 * Clean-up bundle data for bundles that have been successfully received by the DTN peer
 * (all TCP segments have been acknowledged).
 */
void BundleManagerTCP::Cleanup(int nUpToSequenceNo)
{
	Bundle *pLoopBundle = pHead_;
		
	while (pLoopBundle)
	{
		if (pLoopBundle->TCPSequenceNo() <= nUpToSequenceNo)
		{
			pHead_ = pLoopBundle->NextBundle();
			delete pLoopBundle;
			pLoopBundle = pHead_;
		}
		else
			break;
	}
}

/* Remove all bundles from the list. */
void BundleManagerTCP::EmptyList()
{
	Bundle *pLoopBundle = pHead_;
		
	while (pLoopBundle)
	{
		pHead_ = pLoopBundle->NextBundle();
		delete pLoopBundle;
		pLoopBundle = pHead_;
	}
}

/* Returns the number of bundles stored in the manager. */
int BundleManagerTCP::BundlesStored()
{
	int nBundles = 0;

	Bundle *pLoopBundle = pHead_;

	while (pLoopBundle)
	{
		nBundles++;
		pLoopBundle = pLoopBundle->NextBundle();
	}

	return nBundles;
}

/* Returns the bytes stored in the manager. */
int BundleManagerTCP::MemoryAllocated()
{
	int nTotalSize = 0;

	Bundle *pLoopBundle = pHead_;

	while (pLoopBundle)
	{
		nTotalSize += pLoopBundle->HeaderDTN().BundleSize();
		pLoopBundle = pLoopBundle->NextBundle();
	}

	return nTotalSize;	
}

void BundleManagerTCP::DumpBundles()
{
	int nBundles = 0;

	Bundle *pLoopBundle = pHead_;

	while (pLoopBundle)
	{
		nBundles++;
		printf("Bundle %d: ", nBundles);
		pLoopBundle->Dump();
		pLoopBundle = pLoopBundle->NextBundle();
	}
}

int BundleManagerTCP::RemoveBundlesAndReorder(Bundle inBundle, int nMaxSeqNumber)
{
	int nSequence = -1;
	int nRemovedBundles = 0;	

	if (pHead_)
		nSequence = pHead_->TCPSequenceNo();
		
	Bundle *pLoopBundle = pHead_;
	Bundle *pPrevBundle = pHead_;

	bool bLeadingEncountered = false;
	
	while (pLoopBundle)
	{
		/* Make sure we do not cancel segments belonging to in-transit bundles. */
		if (inBundle.IsSameAs(*pLoopBundle) && pLoopBundle->HeaderDTN().LeadingBytes() && pLoopBundle->TCPSequenceNo() > nMaxSeqNumber)
			bLeadingEncountered = true;

		if (bLeadingEncountered && inBundle.IsSameAs(*pLoopBundle) && pLoopBundle->TCPSequenceNo() > nMaxSeqNumber)
		// if (inBundle.IsSameAs(*pLoopBundle) && pLoopBundle->TCPSequenceNo() > nMaxSeqNumber)
		{
			nRemovedBundles++;
			
			if (pLoopBundle == pHead_)
			{
				pHead_ = pLoopBundle->NextBundle();
				pPrevBundle = pHead_;
			}
			else
				pPrevBundle->NextBundle() = pLoopBundle->NextBundle();

			// if (inBundle.HeaderDTN().SequenceNumber() == 86)
			// {
			//	printf("Deleting: ");
			//	pLoopBundle->Dump();
			// }
		
			delete pLoopBundle;
			
			if (pPrevBundle)
				pLoopBundle = pPrevBundle->NextBundle();
			else
				break;
		}
		else
		{
			pPrevBundle = pLoopBundle;
			pLoopBundle = pLoopBundle->NextBundle();
		}
	}
	
	pLoopBundle = pHead_;
	while (pLoopBundle)
	{
		pLoopBundle->TCPSequenceNo() = nSequence;
		nSequence++;
		pLoopBundle = pLoopBundle->NextBundle();
	}
	
	return nRemovedBundles;
}

void BundleManagerTCP::DumpMinMaxSequence()
{
	int minSequence = -1;
	int maxSequence = -1;
	
	if (pHead_)
		minSequence = pHead_->TCPSequenceNo();

	Bundle *pLoopBundle = pHead_;
	while (pLoopBundle)
	{
		maxSequence = pLoopBundle->TCPSequenceNo();
		pLoopBundle = pLoopBundle->NextBundle();
	}
	
	printf("Min sequence: %d, max sequence: %d\n", minSequence, maxSequence);	
}

static class TcpDtnClass : public TclClass {
public:
	TcpDtnClass() : TclClass("Agent/TCP/TCP-DTN") {}
	TclObject* create(int, const char*const*) {
		return (new TcpDtnAgent());
	}
} class_tcpdtn;

TcpDtnAgent::TcpDtnAgent()
{
	m_pDtnAgent = NULL;
	lastSequenceNo_ = 0;
}

TcpDtnAgent::~TcpDtnAgent()
{
}

int TcpDtnAgent::sendBundle(Bundle inBundle, int nBytes)
{
	int nSegments = (nBytes/size_ + (nBytes%size_ ? 1 : 0));
	int nLeftoverBytes	= nBytes;
	// printf("nSegments = %d\n", nSegments);
	for (int i = 1; i <= nSegments; i++)
	{
		
		inBundle.TCPSequenceNo()			= lastSequenceNo_ + i;
		if (nLeftoverBytes > size_)
			inBundle.HeaderDTN().Payload()		= size_;
		else
			inBundle.HeaderDTN().Payload()		= nLeftoverBytes;
		if (inBundle.HeaderDTN().SegmentNumber() == 0)
			inBundle.HeaderDTN().SegmentNumber()		= i;
		// pNewBundleData->m_nNextRVms				= nNextRVms;
		if (inBundle.HeaderDTN().LeadingBytes()	== 1 && i == 1)
			inBundle.HeaderDTN().LeadingBytes() = 1;
		else
			inBundle.HeaderDTN().LeadingBytes() = 0;
			
		bundleManagerTCP_.AddBundle(inBundle);
		nLeftoverBytes -= size_;
		
		//if (inBundle.HeaderDTN().SequenceNumber() == 42)
		//printf("Tcp (%d): adding bundle tail: BundleSeq: %d, BundleSize: %d, TcpSeq: %d, SegNo: %d, StatusRep: %d, LeadBytes: %d\n", 
		//	addr(), inBundle.HeaderDTN().SequenceNumber(), inBundle.HeaderDTN().BundleSize(), inBundle.TCPSequenceNo(), i,
		//	inBundle.HeaderDTN().IsStatusReport(), inBundle.HeaderDTN().LeadingBytes());
	}
	lastSequenceNo_ += nSegments;
	curseq_ += nSegments;
	closed_ = 0;
	send_much(0, 0, maxburst_);
	return nSegments;
}

/* Called by the DtnAgent when pending packets related to a certain bundle must be cancelled from transmitting. */
void TcpDtnAgent::CancelBundleTransmission(Bundle inBundle)
{
	/* Remove packets related to the passed bundle and reorder packet sequence numbers. */
	int nRemovedPackets = bundleManagerTCP_.RemoveBundlesAndReorder(inBundle, maxseq_);
	curseq_ -= nRemovedPackets;
	lastSequenceNo_ -= nRemovedPackets;
	// if (inBundle.HeaderDTN().SequenceNumber() == 86)
	printf("TCP (%d): Cancelling bundle: %d, Offset: %d, BundleSize: %d, Removed: %d\n", 
		addr(), inBundle.HeaderDTN().SequenceNumber(), inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().BundleSize(), nRemovedPackets);
	//bundleManagerTCP_.DumpMinMaxSequence();
}

bool TcpDtnAgent::BundleInTransit(Bundle inBundle)
{
	return (bundleManagerTCP_.FindFirstBundle(inBundle) != NULL);
}
	
void TcpDtnAgent::output_helper(Packet *pPacket)
{
	hdr_tcp *tcph = hdr_tcp::access(pPacket);
	// hdr_cmn *cmnh = hdr_cmn::access(pPacket);
	
	Bundle* pBundle = bundleManagerTCP_.GetBundle(tcph->seqno());
	if (pBundle)
	{
		hdr_dtn *dtnh = hdr_dtn::access(pPacket);
		*dtnh = pBundle->HeaderDTN();
		Bundle* pNextBundle = bundleManagerTCP_.GetBundle(tcph->seqno() + 1);
		/* Let DtnAgent know that a bundle has completed transmitting so that the RTTimer can be set. */
		if (!pNextBundle || !pBundle->IsSameAs(*pNextBundle))
			m_pDtnAgent->BundleTransmitted(*pBundle);
		
		// if (dtnh->SequenceNumber() == 42)
		//printf("Tcp Helper (%d): sending bundle: BundleSeq: %d, TCPSeq: %d, Status: %d, Offset: %d, BundleSize: %d, Leading: %d\n", 
		//	addr(), dtnh->SequenceNumber(), tcph->seqno(), dtnh->IsStatusReport(), dtnh->FragmentOffset(), dtnh->BundleSize(), dtnh->LeadingBytes());	
	}
	//else
	//	printf("TCP Helper (%d): could not find relevant data for: TCPSeq: %d\n", 
	//		addr(), tcph->seqno());	

	TcpAgent::output_helper(pPacket);
}

void TcpDtnAgent::recv_newack_helper(Packet *pPacket) 
{
	TcpAgent::recv_newack_helper(pPacket);
	bundleManagerTCP_.Cleanup(highest_ack_);
}
